import logging
from collections import defaultdict
from datetime import datetime
from ipaddress import IPv4Address, AddressValueError
from typing import Any, Union, Optional

from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass

from ipfabric.tools.shared import date_parser

logger = logging.getLogger("ipfabric")
DEFAULT_SNAPSHOT = "$last"
NAMED_SNAPSHOTS = [DEFAULT_SNAPSHOT, "$prev", "$first"]


class Config(BaseModel):
    config_id: str = Field(None, alias="id")
    sn: str
    hostname: str
    config_hash: str = Field(None, alias="hash")
    status: str
    last_change: datetime = Field(None, alias="lastChangeAt")
    last_check: datetime = Field(None, alias="lastCheckAt")
    text: Optional[str] = None

    def get_text_config(self, client: Any, sanitized: bool = True) -> str:
        """gets a device config
        Args:
            client: IPF Client
            sanitized: bool to determine to return sensitive data

        Returns:
            string containing config of a device
        """
        res = client.get(
            "/tables/management/configuration/download",
            params=dict(hash=self.config_hash, sanitized=sanitized),
        )
        res.raise_for_status()
        self.text = res.text
        return res.text


@dataclass
class DeviceConfigs:
    client: Any

    def get_all_configurations(self, device: Optional[str] = None, sn: Optional[str] = None) -> Union[dict, None]:
        """Get all configurations in IP Fabric

        Args:
            device: Hostname (case insensitive) filter
            sn: Serial number of device
        Returns:
            results: dict: {sn: [Config, Config]}
        """
        if device or sn:
            filters = dict(sn=["eq", sn]) if sn else dict(hostname=["ieq", device])
            res = self.client.fetch_all(
                "tables/management/configuration",
                sort={"order": "desc", "column": "lastChangeAt"},
                columns=[
                    "id",
                    "sn",
                    "hostname",
                    "lastChangeAt",
                    "lastCheckAt",
                    "status",
                    "hash",
                ],
                filters=filters,
                snapshot=False,
            )
            if len(res) == 0:
                logger.warning(f"Could not find any configurations for device '{device}'.")
                return None
        else:
            res = self.client.fetch_all(
                "tables/management/configuration",
                sort={"order": "desc", "column": "lastChangeAt"},
                columns=[
                    "id",
                    "sn",
                    "hostname",
                    "lastChangeAt",
                    "lastCheckAt",
                    "status",
                    "hash",
                ],
                snapshot=False,
            )
        results = defaultdict(list)
        [results[cfg["sn"]].append(Config(**cfg)) for cfg in res]
        return results

    def _search_ip(self, ip: str, snapshot_id: str = None, log: bool = False) -> dict:
        res = self.client.fetch_all(
            "tables/addressing/managed-devs",
            columns=["ip", "hostname", "sn"],
            reports="/technology/addressing/managed-ip",
            filters=dict(ip=["eq", ip]),
        )
        if len(res) == 1 and not log:
            return {"hostname": res[0]["hostname"], "sn": res[0]["sn"]}
        if len(res) == 1 and log:
            res = self.client.inventory.devices.all(
                columns=["hostname", "taskKey", "sn"],
                snapshot_id=snapshot_id,
                filters=dict(sn=["eq", res[0]["sn"]]),
            )
            return {"hostname": res[0]["hostname"], "taskKey": res[0]["taskKey"], "sn": res[0]["sn"]}
        elif len(res) > 1:
            logger.warning(f"Found multiple entries for IP '{ip}'.")
        elif len(res) == 0:
            logger.warning(f"Could not find a matching IP for '{ip}'.")
        return {"hostname": None, "sn": None}

    def get_configuration(
        self, device: str = None, sn: str = None, sanitized: bool = True, date: Union[str, tuple] = DEFAULT_SNAPSHOT
    ) -> Union[Config, None]:
        """Gets last configuration of a device based on hostname or IP or IPF Unique Serial Number

        Args:
        device: Hostname or IP
        sn: Serial Number
        sanitized: Default True to mask passwords
        date: Defaults to latest config. Values in [$last, $prev, $first] or can be a
            tuple of a date range to get the latest snapshot in that range.
            Date can be string or int in seconds ("11/22/ 1:30", 1637629200)
        Returns:
            Returns a result or None
        """
        if not isinstance(date, tuple) and date not in NAMED_SNAPSHOTS:
            raise SyntaxError(f"Date must be in {NAMED_SNAPSHOTS} or tuple ('startDate', 'endDate')")
        if not sn:
            sn = self._validate_device(device)["sn"]
            if not sn:
                return None
        cfgs = self.get_all_configurations(sn=sn)
        if not cfgs:
            return None
        cfg = self._get_hash(cfgs[sn], date)
        if cfg:
            return self.get_text_config(cfg, sanitized)
        else:
            logger.error(f"Could not find a configuration with date {date}")
            return None

    def get_text_config(self, cfg: Config, sanitized: bool = True) -> Config:
        """gets a devices config
        Args:
            cfg: Config from get_configuration method
            sanitized: bool to determine to return sensitive data

        Returns:
            string containing config of a device
        """
        res = self.client.get(
            "/tables/management/configuration/download",
            params=dict(hash=cfg.config_hash, sanitized=sanitized),
        )
        res.raise_for_status()
        cfg.text = res.text
        return cfg

    @staticmethod
    def _get_hash(configs, date):
        if isinstance(date, tuple):
            start = date_parser(date[0])
            end = date_parser(date[1])
            for cfg in configs:
                if start < cfg.last_change < end:
                    return cfg
        elif date == DEFAULT_SNAPSHOT:
            return configs[0]
        elif date == "$prev" and len(configs) > 1:
            return configs[1]
        elif date == "$first":
            return configs[-1]
        return None

    def _validate_device(self, device: str, snapshot_id: str = None, log: bool = False) -> dict:
        try:
            ip = IPv4Address(device)
            return self._search_ip(str(ip), snapshot_id=snapshot_id, log=log)
        except AddressValueError:
            res = self.client.inventory.devices.all(
                columns=["hostname", "taskKey", "sn"],
                filters=dict(hostname=["ieq", device]),
                snapshot_id=snapshot_id,
            )
            if len(res) == 1:
                return {"hostname": res[0]["hostname"], "taskKey": res[0]["taskKey"], "sn": res[0]["sn"]}
            elif len(res) == 0:
                logger.warning(f"Could not find a matching device for '{device}'")
            elif len(res) > 1:
                logger.warning(f"Found multiple devices matching '{device}'.")
        return {"hostname": None, "sn": None}

    def get_log(self, device: str, snapshot_id: str = None):
        device = self._validate_device(device, snapshot_id=snapshot_id, log=True)
        if not device["sn"]:
            return None
        return self.get_text_log(device)

    def get_text_log(self, device: dict):
        res = self.client.get("/os/logs/task/" + device["taskKey"])
        res.raise_for_status()
        return res.text



from .graphs import IPFDiagram, Diagram
from .input_models import (
    Unicast,
    Multicast,
    Host2GW,
    Network,
    OtherOptions,
    Algorithm,
    EntryPoint,
    Layout,
    NetworkSettings,
    PathLookupSettings,
    Overlay,
    VALID_NET_PROTOCOLS,
    VALID_PATH_PROTOCOLS,
)

__all__ = [
    "IPFDiagram",
    "Diagram",
    "Unicast",
    "Multicast",
    "Host2GW",
    "Network",
    "OtherOptions",
    "Algorithm",
    "Overlay",
    "icmp",
    "NetworkSettings",
    "PathLookupSettings",
    "EntryPoint",
    "VALID_NET_PROTOCOLS",
    "VALID_PATH_PROTOCOLS",
    "Layout",
]



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Sdn(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def aci_endpoints(self):
        return models.Table(client=self.client, endpoint="tables/aci/endpoints", sn=self.sn)

    @property
    def aci_vlan(self):
        return models.Table(client=self.client, endpoint="tables/aci/vlan", sn=self.sn)

    @property
    def aci_vrf(self):
        return models.Table(client=self.client, endpoint="tables/aci/vrf", sn=self.sn)

    @property
    def aci_dtep(self):
        return models.Table(client=self.client, endpoint="tables/aci/dtep", sn=self.sn)

    @property
    def vxlan_vtep(self):
        return models.Table(client=self.client, endpoint="tables/vxlan/vtep", sn=self.sn)

    @property
    def vxlan_peers(self):
        return models.Table(client=self.client, endpoint="tables/vxlan/peers", sn=self.sn)

    @property
    def vxlan_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/vxlan/interfaces", sn=self.sn)

    @property
    def vxlan_vni(self):
        return models.Table(client=self.client, endpoint="tables/vxlan/vni", sn=self.sn)

    @property
    def apic_controllers(self):
        return models.Table(client=self.client, endpoint="tables/apic/controllers", sn=self.sn)

    @property
    def apic_contexts(self):
        return models.Table(client=self.client, endpoint="tables/apic/contexts", sn=self.sn)

    @property
    def apic_bridge_domains(self):
        return models.Table(client=self.client, endpoint="tables/apic/bridge-domains", sn=self.sn)

    @property
    def apic_applications(self):
        return models.Table(client=self.client, endpoint="tables/apic/applications", sn=self.sn)

    @property
    def apic_endpoint_groups(self):
        return models.Table(client=self.client, endpoint="tables/apic/endpoint-groups", sn=self.sn)

    @property
    def apic_endpoint_groups_contracts(self):
        return models.Table(client=self.client, endpoint="tables/apic/endpoint-groups/contracts", sn=self.sn)

    @property
    def apic_contracts(self):
        return models.Table(client=self.client, endpoint="tables/apic/contracts", sn=self.sn)

    @property
    def apic_service_graphs(self):
        return models.Table(client=self.client, endpoint="tables/apic/service-graphs", sn=self.sn)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Stp(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def bridges(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/bridges", sn=self.sn)

    @property
    def instances(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/instances", sn=self.sn)

    @property
    def vlans(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/vlans", sn=self.sn)

    @property
    def ports(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/ports", sn=self.sn)

    @property
    def neighbors(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/neighbors", sn=self.sn)

    @property
    def guards(self):
        return models.Table(client=self.client, endpoint="tables/spanning-tree/guards", sn=self.sn)

    @property
    def inconsistencies(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/inconsistencies/summary", sn=self.sn)

    @property
    def inconsistencies_details(self):
        return models.Table(client=self.client, endpoint="tables/interfaces/inconsistencies/details", sn=self.sn)

    @property
    def inconsistencies_ports_vlan_mismatch(self):
        return models.Table(
            client=self.client, endpoint="tables/spanning-tree/inconsistencies/neighbor-ports-vlan-mismatch", sn=self.sn
        )

    @property
    def inconsistencies_ports_multiple_neighbors(self):
        return models.Table(
            client=self.client, endpoint="tables/spanning-tree/inconsistencies/ports-multiple-neighbors", sn=self.sn
        )

    @property
    def inconsistencies_stp_cdp_ports_mismatch(self):
        return models.Table(
            client=self.client, endpoint="tables/spanning-tree/inconsistencies/stp-cdp-ports-mismatch", sn=self.sn
        )

    @property
    def inconsistencies_multiple_stp(self):
        return models.Table(
            client=self.client, endpoint="tables/spanning-tree/inconsistencies/multiple-stp", sn=self.sn
        )



from .configuration import DeviceConfigs
from .discovery_history import DiscoveryHistory
from .restore_intents import RestoreIntents
from .shared import parse_mac, TIMEZONES, VALID_REFS, validate_ip_network_str, VALID_IP
from .site_seperation_report import map_devices_to_rules
from .vulnerabilities import Vulnerabilities

__all__ = [
    "DeviceConfigs",
    "Vulnerabilities",
    "RestoreIntents",
    "DiscoveryHistory",
    "map_devices_to_rules",
    "parse_mac",
    "TIMEZONES",
    "VALID_REFS",
    "validate_ip_network_str",
    "VALID_IP",
]



import logging
from warnings import warn

logger = logging.getLogger("ipfabric")


def deprecated_args_decorator(version: str, arg_type=None, no_args: bool = True):  # noqa: C901

    def inner(func):

        fname = func.__name__

        def wrapper(*args, **kwargs):
            if arg_type:
                arg_types = [arg_type] if isinstance(arg_type, str) else arg_type
                msg = (
                    f"`{fname}()` no longer accepts parameter of `{arg_type}` and will be removed in version {version}."
                )
                if args:
                    for arg in args:
                        if type(arg).__name__ in arg_types:
                            warn(msg, DeprecationWarning, stacklevel=2)
                            logger.warning(msg)
                if kwargs:
                    for arg in kwargs.values():
                        if type(arg).__name__ in arg_types:
                            warn(msg, DeprecationWarning, stacklevel=2)
                            logger.warning(msg)
            if no_args and (len(args) > 1 or kwargs):
                warn(
                    f"`{fname}()` parameters are deprecated and will be removed in version {version}.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                logger.warning(f"`{fname}()` parameters are deprecated and will be removed in version {version}.")
            return func(*args, **kwargs)

        return wrapper

    return inner



import logging
from typing import Any, ClassVar, Union

from pydantic import ValidationError
from pydantic.dataclasses import dataclass

from ipfabric.settings.vendor_api_models import (
    AWS,
    Azure,
    CheckPointApiKey,
    CheckPointUserAuth,
    CiscoAPIC,
    CiscoFMC,
    ForcePoint,
    GCP,
    JuniperMist,
    Merakiv1,
    Prisma,
    RuckusVirtualSmartZone,
    SilverPeak,
    Versa,
    Viptela,
    NSXT,
)

API_MODELS = (
    AWS,
    Azure,
    CheckPointApiKey,
    CheckPointUserAuth,
    CiscoAPIC,
    CiscoFMC,
    ForcePoint,
    GCP,
    JuniperMist,
    Merakiv1,
    Prisma,
    RuckusVirtualSmartZone,
    SilverPeak,
    Versa,
    Viptela,
    NSXT,
)

TYPE_TO_MODEL = {c.model_fields["type"].default: c for c in API_MODELS}

CONNECTION_PARAMS = [
    "rejectUnauthorized",
    "respectSystemProxyConfiguration",
    "maxConcurrentRequests",
    "maxCapacity",
    "refillRate",
    "refillRateIntervalMs",
]

logger = logging.getLogger("ipfabric")


@dataclass
class VendorAPI:
    client: Any
    _api_url: ClassVar[str] = "settings/vendor-api"

    def get_vendor_apis(self) -> list:
        """
        Get all vendor apis and sets them in the Authentication.apis
        :return: self.credentials
        """
        res = self.client.get(self._api_url)
        res.raise_for_status()
        return res.json()

    def add_vendor_api(self, api: Union[API_MODELS]) -> dict:
        params = api.model_dump()
        res = self.client.post(self._api_url, json=params)
        res.raise_for_status()
        return res.json()

    @staticmethod
    def _return_api_id(api_id: Union[dict, str, int]) -> str:
        if isinstance(api_id, dict):
            api_id = api_id["id"]
        elif isinstance(api_id, int):
            api_id = str(api_id)
        return api_id

    def delete_vendor_api(self, api_id: Union[dict, str, int]):
        api_id = self._return_api_id(api_id)
        res = self.client.delete(self._api_url + "/" + api_id)
        res.raise_for_status()
        return res.status_code

    def _enable_api(self, api_id: Union[dict, str, int], enable: bool = True) -> int:
        api_id = self._return_api_id(api_id)
        res = self.client.patch(self._api_url + "/" + api_id, json={"isEnabled": enable})
        res.raise_for_status()
        return res.status_code

    def enable_vendor_api(self, api_id: Union[dict, str, int]) -> int:
        return self._enable_api(api_id)

    def disable_vendor_api(self, api_id: Union[dict, str, int]) -> int:
        return self._enable_api(api_id, False)

    @staticmethod
    def _validate_update(params):
        if params["type"] == "checkpoint-mgmt-api":
            try:
                CheckPointApiKey.model_validate(params)
            except ValidationError:
                try:
                    CheckPointUserAuth.model_validate(params)
                except ValidationError:
                    raise SyntaxError(
                        "Error validating updated data for Checkpoint API "
                        "please provide either apiKey or username and password."
                    )
        else:
            try:
                TYPE_TO_MODEL[params["type"]].model_validate(params)
            except ValidationError:
                raise SyntaxError("Error validating updated data for Vendor API please see missing required arguments.")
        return params

    def update_vendor_api(
        self, current: dict, update: Union[Union[API_MODELS], dict], restore_conn_defaults: bool = False
    ) -> dict:
        current.pop("details", None)
        api_id = current.pop("id")

        if isinstance(update, API_MODELS):
            update = vars(update)

        params = {**current, **update}
        if restore_conn_defaults:
            default = {
                k: v for k, v in vars(TYPE_TO_MODEL[params["type"]].model_construct()).items() if k in CONNECTION_PARAMS
            }
            params.update(default)

        self._validate_update(params)

        res = self.client.put(self._api_url + "/" + str(api_id), json=params)
        res.raise_for_status()
        return res.json()



import logging
from typing import Optional, Dict, List, Any, Union, overload, Literal

import deepdiff
from pydantic import BaseModel

from ipfabric.models import Devices
from ipfabric.tools.shared import create_filter

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

logger = logging.getLogger("ipfabric")

IGNORE_COLUMNS = {"id"}
EXPORT_FORMAT = Literal["json", "csv", "df"]
DEVICE_EXPORT_FORMAT = Literal["json", "csv", "object", "df"]


class BaseTable(BaseModel):
    """model for table data"""

    endpoint: str
    client: Any = None

    @property
    def name(self):
        return self.endpoint.split("/")[-1]

    @property
    def abs_endpoint(self):
        return self.endpoint if self.endpoint[0] == "/" else "/" + self.endpoint

    @overload
    def fetch(
        self,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> List[dict]: ...

    @overload
    def fetch(
        self,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> DataFrame: ...

    def fetch(
        self,
        export: EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters'
            attr_filters: dict: Optional dictionary of Attribute filters
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            limit: Default to 1,000 rows
            start: Starts at 0
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pd.DataFrame]: List of dict if json, bytes string if CSV, DataFrame if df
        """
        return self.client.fetch(
            self.endpoint,
            export=export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            sort=sort,
            limit=limit,
            start=start,
            snapshot=False,
            csv_tz=csv_tz,
        )

    @overload
    def all(
        self,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        sort: Optional[dict] = None,
    ) -> List[dict]: ...

    @overload
    def all(
        self,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def all(
        self,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        sort: Optional[dict] = None,
    ) -> DataFrame: ...

    def all(
        self,
        export: EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters
            attr_filters: dict: Optional dictionary of Attribute filters
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pd.DataFrame]: List of dict if json, bytes string if CSV, DataFrame if df
        """
        return self.client.fetch_all(
            self.endpoint,
            export=export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            sort=sort,
            snapshot=False,
            csv_tz=csv_tz,
        )

    def count(
        self,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
    ) -> int:
        """
        Gets count of table
        :param filters: dict: Optional filters
        :param attr_filters: dict: Optional dictionary of Attribute filters
        :return: int: Count
        """
        return self.client.get_count(self.endpoint, filters=filters, attr_filters=attr_filters, snapshot=False)


class Table(BaseTable, BaseModel):
    """model for table data"""

    sn: Optional[str] = None

    def _dev_filters(self, url, columns=None, filters=None):
        if self.sn:
            all_columns, f = create_filter(self.client, url, filters=filters, sn=self.sn)
            return columns or all_columns, f
        else:
            return columns, filters

    @overload
    def fetch(
        self,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> List[dict]: ...

    @overload
    def fetch(
        self,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> DataFrame: ...

    def fetch(
        self,
        export: EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters'
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            limit: Default to 1,000 rows
            start: Starts at 0
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pd.DataFrame]: List of dict if json, bytes string if CSV, DataFrame if df
        """
        columns, filters = self._dev_filters(self.endpoint, columns, filters)
        return self.client.fetch(
            self.endpoint,
            export=export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports,
            sort=sort,
            limit=limit,
            start=start,
            csv_tz=csv_tz,
        )

    @overload
    def all(
        self,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
    ) -> List[dict]: ...

    @overload
    def all(
        self,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def all(
        self,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
    ) -> DataFrame: ...

    def all(
        self,
        export: EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], bytes, pd.DataFrame]: List of dict if json, bytes string if CSV, DataFrame if df
        """
        columns, filters = self._dev_filters(self.endpoint, columns, filters)
        return self.client.fetch_all(
            self.endpoint,
            export=export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports,
            sort=sort,
            csv_tz=csv_tz,
        )

    def count(
        self,
        filters: Optional[dict] = None,
        snapshot_id: Optional[str] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
    ) -> int:
        """
        Gets count of table
        Args:
            filters: dict: Optional filters
            snapshot_id: str: Optional snapshot ID to override class
            attr_filters: dict: Optional dictionary of Attribute filters

        Returns:
            int: Count of rows
        """
        _, filters = self._dev_filters(self.endpoint, filters=filters)
        return self.client.get_count(self.endpoint, filters=filters, attr_filters=attr_filters, snapshot_id=snapshot_id)

    @staticmethod
    def _ignore_columns(columns: set, columns_ignore: set):
        """
        Determines which columns to use in the query.
        Args:
            columns: set : Set of columns to use
            columns_ignore: set : Set of columns to ignore

        Returns:
            list[str]: List of columns to use
        """
        cols_for_return = set()
        for col in columns:
            if col in columns_ignore and col != "id":
                logger.debug(f"Column {col} in columns_ignore, ignoring")
                continue
            cols_for_return.add(col)
        return cols_for_return

    def _nested_columns(self, table_columns: set, nested_keys: set):
        # get any nested columns
        nested_cols_for_return = None
        # if the user passed nested_column_returns
        # if nested_keys:
        #     # get a row of data
        #     one_row_of_data = self.fetch(limit=1)
        #     # loop over each column in the table columns
        #     for col in table_columns:
        #         # check if the data returned has nested columns
        #         if isinstance(one_row_of_data[0][col], list):
        #             logger.debug(f"Column: {col}, one row of data: {one_row_of_data}")
        #             check_len_of_data = len(one_row_of_data[0][col])
        #             if check_len_of_data >= 0:
        #                 logger.warning("Nested Column was Detected, but no data to determine nested columns, skipping.")
        #                 continue
        #             # get all the nested column names
        #             nested_cols = set(one_row_of_data[0][col][0].keys())
        #             # create a set of nested column names requested by the user
        #             nested_cols_for_return = {col for col in nested_cols if col in nested_keys}
        if nested_keys:
            nested_cols = set(self.oas[self.endpoint].nested_columns)
            nested_cols_for_return = {col for col in nested_cols if col in nested_keys}
        return nested_cols_for_return

    def _compare_determine_columns(self, columns: set, columns_ignore: set, unique_keys: set, nested_keys: set):
        """
        Determines which columns to use in the query, and which columns to use
        Only supports a single nested column
        Args:
            columns: set : Set of columns to use
            columns_ignore: set : Set of columns to ignore
            unique_keys: set : Set of columns for unique keys
            nested_keys: set: Set of nested columns that belongs to a column in columns, columns_ignore, or unique_keys

        Returns:
            tuple[list, list]: List of columns to use in query, List of columns to use when sorting data
        """
        # get all columns for the table
        table_columns = set(self.client.get_columns(self.endpoint))
        nested_cols_for_return = self._nested_columns(table_columns, nested_keys)

        # Must always ignore some columns
        columns_ignore.update(IGNORE_COLUMNS)

        cols_for_return = set()
        # user passes unique_keys
        if unique_keys:
            if not table_columns.issuperset(unique_keys):
                raise ValueError(f"Unique Key(s) {unique_keys - table_columns} not in table {self.name}")
            for u in unique_keys:
                cols_for_return.add(u)
            return list(cols_for_return), nested_cols_for_return
        # user passes columns
        if columns:
            if not table_columns.issuperset(columns):
                raise ValueError(f"Column(s) {columns - table_columns} not in table {self.name}")
            cols_for_return.update(self._ignore_columns(columns, columns_ignore))
            return list(cols_for_return), nested_cols_for_return
        # user does not pass columns
        # or only passed columns_ignore
        else:
            cols_for_return.update(self._ignore_columns(table_columns, columns_ignore))
            return list(cols_for_return), nested_cols_for_return

    @staticmethod
    def _remove_keys_from_dict(dict_object, keys_to_remove):
        if not keys_to_remove:
            return dict_object

        for v in dict_object.values():
            if not isinstance(v, list):
                continue

            for item in v:
                for key in keys_to_remove:
                    if key in item:
                        item.pop(key)

        return dict_object

    def _hash_data(self, json_data, unique_keys=None, nested_columns_ignore=None):
        """
        Hashes data. Turns any data into a string and hashes it, then returns the hash as a key for the data
        Args:
            json_data: list[dict] : List of dictionaries to hash
            unique_keys: list[str] : List of keys to use for hashing
            nested_columns_ignore: list[str]: List of nested column keys to filter data

        Returns:
            dict[str]: dictionary with hash as key and values as the original data
        """
        # loop over each obj, turn the obj into a string, and hash it
        return_json = dict()
        # user passes unique_keys
        if unique_keys:
            # loop over each response
            for dict_obj in json_data:
                dict_for_hash = self._remove_keys_from_dict(dict_object=dict_obj, keys_to_remove=nested_columns_ignore)
                # create a dictionary, of only the keys/columns requested
                hash_key = {key: dict_for_hash[key] for key in unique_keys}
                # hash the data
                unique_hash = deepdiff.DeepHash(hash_key)[hash_key]
                # check if the data has already been processed.
                if unique_hash in return_json:
                    # raise error if data has already been processed
                    raise KeyError(f"Unique Key(s) {unique_keys} are not unique, please adjust unique_keys input.")
                # store the data, using the hash as a key
                return_json[unique_hash] = dict_for_hash
        # user passes columns and/or column ignore
        else:
            # loop over each response
            for dict_obj in json_data:
                dict_for_hash = self._remove_keys_from_dict(dict_object=dict_obj, keys_to_remove=nested_columns_ignore)
                # hash each object, and store it in a dictionary, with the hash as the key
                return_json[deepdiff.DeepHash(dict_for_hash)[dict_for_hash]] = dict_for_hash
        return return_json

    @staticmethod
    def _make_set(data: Union[list, set, str] = None):
        if isinstance(data, str):
            return {data}
        elif data is None:
            return set()
        else:
            return set(data)

    def compare(
        self,
        snapshot_id: str = None,
        columns: Union[list, set] = None,
        columns_ignore: Union[list, set, str] = None,
        unique_keys: Union[list, set, str] = None,
        nested_columns_ignore: Union[list, set, str] = None,
        **kwargs,
    ):
        """
        Compares a table from the current snapshot to the snapshot_id passed.
        Args:
            snapshot_id: str : The snapshot_id to compare to.
            columns: list : List of columns to compare. If None, will compare all columns.
            columns_ignore: list : List of columns to ignore. If None, will always ignore 'id' column.
            unique_keys: list : List of columns to use as unique keys. If None, will use all columns as primary key.
            nested_columns_ignore: List of columns that belong to a nested columns
            **kwargs: dict : Optional Table.all() arguments to apply to the table before comparing.

        Returns:
            dict : dictionary containing the differences between the two snapshots.
                   Possible keys are 'added', 'removed' and 'changed'.
        """
        return_dict = dict()

        # determine which columns to use in query
        columns = self._make_set(columns)
        columns_ignore = self._make_set(columns_ignore)
        unique_keys = self._make_set(unique_keys)
        nested_columns_ignore = self._make_set(nested_columns_ignore)
        cols_for_query, nested_cols = self._compare_determine_columns(
            columns, columns_ignore, unique_keys, nested_columns_ignore
        )
        data = self.all(columns=cols_for_query, **kwargs)
        data_compare = self.all(snapshot_id=snapshot_id, columns=cols_for_query, **kwargs)

        # since we turned the values into a hash, we can just compare the keys
        if unique_keys:
            hashed_data_unique = self._hash_data(data, unique_keys, nested_cols)
            hashed_data_compare_unique = self._hash_data(data_compare, unique_keys, nested_cols)
            changed = [
                hashed_data_unique[hashed_str]
                for hashed_str in hashed_data_unique.keys()
                if hashed_str not in hashed_data_compare_unique.keys()
            ]
            return_dict["changed"] = changed
            return return_dict
        # compare both ways
        hashed_data = self._hash_data(data, nested_columns_ignore=nested_cols)
        hashed_data_compare = self._hash_data(data_compare, nested_columns_ignore=nested_cols)
        added = [
            hashed_data[hashed_str] for hashed_str in hashed_data.keys() if hashed_str not in hashed_data_compare.keys()
        ]
        removed = [
            hashed_data_compare[hashed_str]
            for hashed_str in hashed_data_compare.keys()
            if hashed_str not in hashed_data.keys()
        ]
        return_dict["added"] = added
        return_dict["removed"] = removed
        return return_dict

    def join_table(self, on_column: str, join_table_path: str) -> Dict:
        """
        performs a join operation on on_columns, assuming on_columns are columns that
        exist in both self and join_table_path. If on_columns are not columns in
        set(self.client.get_columns(self.endpoint)), ValueError is raised.

        Args:
            on_column: str : name of column to use for join operation
            join_table_path: URL of table to use for join operation

        Returns: dict: Dictionary of result of joined table.
         possible keys <join_table_path>, self.endpoint, joined, no_matches
        """
        self_cols = set(self.client.get_columns(self.endpoint))
        if not {on_column}.issubset(self_cols):
            raise ValueError(f"Column {on_column}, is not subset of {self_cols}.")
        self_table = self.client.fetch_all(self.endpoint)
        join_table = self.client.fetch_all(join_table_path)

        joined_data = [
            {**self_dict, **join_dict}
            for self_dict in self_table
            for join_dict in join_table
            if self_dict[on_column] == join_dict[on_column]
        ]

        no_matches = [
            join_dict
            for join_dict in join_table
            if join_dict[on_column] not in (entry[on_column] for entry in joined_data)
        ]

        dict_for_return = {
            "joined": joined_data,
            f"{self.endpoint}": self_table,
            f"{join_table_path}": join_table,
            "no_matches": no_matches,
        }
        return dict_for_return


class DeviceTable(Table):
    """model for Device Table data"""

    endpoint: str = "tables/inventory/devices"

    def _as_model(self, devices, export):
        if export != "object":
            return devices
        return Devices(
            client=self.client,
            snapshot_id=self.client.snapshot_id,
            devices=devices,
        )

    @overload
    def fetch(
        self,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> List[dict]: ...

    @overload
    def fetch(
        self,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def fetch(
        self,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> DataFrame: ...

    @overload
    def fetch(
        self,
        export: Literal["object"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
    ) -> Devices: ...

    def fetch(
        self,
        export: DEVICE_EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        limit: Optional[int] = 1000,
        start: Optional[int] = 0,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv, object]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters'
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            limit: Default to 1,000 rows
            start: Starts at 0
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
            Union[List[dict], Devices, bytes, DataFrame]: Default List of dicts 'json', Devices object if 'object',
                                                          bytes if 'csv', pandas.DataFrame if 'df'
        """
        devices = super(DeviceTable, self).fetch(
            export="json" if export in ["json", "object"] else export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports if export not in ["csv", "object"] else None,
            sort=sort,
            limit=limit,
            start=start,
            csv_tz=csv_tz,
        )
        return self._as_model(devices, export)

    @overload
    def all(
        self,
        export: Literal["json"] = ...,
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
    ) -> List[dict]: ...

    @overload
    def all(
        self,
        export: Literal["csv"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ) -> bytes: ...

    @overload
    def all(
        self,
        export: Literal["df"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
    ) -> DataFrame: ...

    @overload
    def all(
        self,
        export: Literal["object"],
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        sort: Optional[dict] = None,
        client: Optional[Any] = None,
    ) -> Devices: ...

    def all(
        self,
        export: DEVICE_EXPORT_FORMAT = "json",
        columns: Optional[list] = None,
        filters: Optional[dict] = None,
        attr_filters: Optional[Dict[str, List[str]]] = None,
        snapshot_id: Optional[str] = None,
        reports: Optional[Union[bool, list, str]] = False,
        sort: Optional[dict] = None,
        csv_tz: Optional[str] = None,
    ):
        """Gets all data from corresponding endpoint

        Args:
            export: str: Export format to return [json, csv, object]; default is json.
            columns: Optional columns to return, default is all
            filters: Optional filters
            attr_filters: dict: Optional dictionary of Attribute filters
            snapshot_id: Optional snapshot ID to override class
            reports: True to return Intent Rules (also accepts string of frontend URL) or a list of report IDs
            sort: Dictionary to apply sorting: {"order": "desc", "column": "lastChange"}
            csv_tz: str: Default None, set a timezone to return human-readable dates when using CSV;
                         see `ipfabric.tools.shared.TIMEZONES`
        Returns:
             Union[List[dict], Devices, bytes, DataFrame]: Default List of dicts 'json', Devices object if 'object',
                                                           bytes if 'csv', pandas.DataFrame if 'df'
        """
        devices = super(DeviceTable, self).all(
            export="json" if export in ["json", "object"] else export,
            columns=columns,
            filters=filters,
            attr_filters=attr_filters,
            snapshot_id=snapshot_id,
            reports=reports if export not in ["csv", "object"] else None,
            sort=sort,
            csv_tz=csv_tz,
        )
        return self._as_model(devices, export)



import logging
from copy import deepcopy
from typing import Optional

from pydantic.dataclasses import dataclass

from ipfabric.settings.attributes import Attributes
from ipfabric.settings.site_separation import SiteSeparation

logger = logging.getLogger("ipfabric")


@dataclass
class Matches:
    hostname: str
    sn: str
    old_site_name: str
    rule_type: str
    new_site_name: Optional[str] = None
    rule_number: Optional[int] = None
    rule_note: Optional[str] = None
    regex: Optional[str] = None
    transformation: Optional[str] = None


def check_attributes(ipf, devices):
    matches = list()
    attributes = {a["sn"]: a for a in Attributes(ipf).all(filters={"name": ["eq", "siteName"]})}
    for sn, dev in deepcopy(devices).items():
        if sn in attributes:
            matches.append(
                Matches(
                    hostname=dev["hostname"],
                    sn=sn,
                    old_site_name=attributes[sn]["value"],
                    rule_type="attributes",
                    rule_number=-1,
                )
            )
            devices.pop(sn)
    return matches


def _create_device_match(match, data, rule, devices, idx):
    site = match["siteName"] if data["matchingGroupApplied"] else rule["siteName"]
    transformation = rule["transformation"] if rule["transformation"] != "none" else None
    return Matches(
        hostname=match["hostname"],
        old_site_name=devices[match["sn"]]["siteName"],
        sn=match["sn"],
        new_site_name=site,
        rule_type=rule["type"],
        rule_number=idx,
        rule_note=rule["note"],
        regex=rule["regex"],
        transformation=transformation,
    )


def map_devices_to_rules(ipf, snapshot_id: str = "$last"):
    ss = SiteSeparation(ipf)
    devices = {
        d["sn"]: d for d in ipf.inventory.devices.all(columns=["hostname", "sn", "siteName"], snapshot_id=snapshot_id)
    }
    rules = ss.get_separation_rules()

    matches = check_attributes(ipf, devices) if rules["manualEnabled"] else list()

    for idx, rule in enumerate(rules["rules"]):
        data = ss.get_rule_matches(rule)
        for match in data["matched"]:
            if match["sn"] in devices:
                matches.append(_create_device_match(match, data, rule, devices, idx))
                devices.pop(match["sn"], None)

    for sn, dev in devices.items():
        matches.append(
            Matches(hostname=dev["hostname"], sn=sn, old_site_name=dev["siteName"], rule_type="noMatchingRule")
        )
    return [vars(m) for m in matches]



import logging
from copy import deepcopy
from typing import Any

from pydantic.dataclasses import dataclass

logger = logging.getLogger("ipfabric")


@dataclass
class SiteSeparation:
    ipf: Any

    def get_separation_rules(self):
        return self.ipf.get("settings/site-separation").json()

    def _post_rule(self, data):
        resp = self.ipf.post("settings/site-separation/test-regex", json=data)
        resp.raise_for_status()
        return resp.json()

    @staticmethod
    def _create_rule(transformation, regex):
        transformation = transformation.lower()
        if transformation not in ["uppercase", "lowercase", "none"]:
            raise SyntaxError('Transformation type is not in ["uppercase", "lowercase", "none"].')
        return {"regex": regex, "transformation": transformation}

    def get_rule_matches(self, rule):
        rule = deepcopy(rule)
        [rule.pop(key, None) for key in ["id", "note", "siteName"]]
        return self._post_rule(rule)

    def get_hostname_matches(self, regex, transformation):
        rule = self._create_rule(transformation, regex)
        rule["type"] = "regexHostname"
        return self._post_rule(rule)

    def get_snmp_matches(self, regex, transformation):
        rule = self._create_rule(transformation, regex)
        rule["type"] = "regexSnmpLocation"
        return self._post_rule(rule)



import logging
from typing import Any, Optional

from pydantic import BaseModel

from ipfabric import models

logger = logging.getLogger("ipfabric")


class Platforms(BaseModel):
    client: Any = None
    sn: Optional[str] = None

    @property
    def environment_power_supplies(self):
        return models.Table(client=self.client, endpoint="tables/inventory/power-supplies", sn=self.sn)

    @property
    def environment_power_supplies_fans(self):
        return models.Table(client=self.client, endpoint="tables/inventory/power-supplies-fans", sn=self.sn)

    @property
    def environment_fans(self):
        return models.Table(client=self.client, endpoint="tables/inventory/fans", sn=self.sn)

    @property
    def environment_modules(self):
        return models.Table(client=self.client, endpoint="tables/inventory/modules", sn=self.sn)

    @property
    def environment_temperature_sensors(self):
        return models.Table(client=self.client, endpoint="tables/inventory/temperature-sensors", sn=self.sn)

    @property
    def cisco_fabric_path_summary(self):
        return models.Table(client=self.client, endpoint="tables/platforms/fabric-path/summary", sn=self.sn)

    @property
    def cisco_fabric_path_isis_neighbors(self):
        return models.Table(client=self.client, endpoint="tables/platforms/fabric-path/neighbors", sn=self.sn)

    @property
    def cisco_fabric_path_switches(self):
        return models.Table(client=self.client, endpoint="tables/platforms/fabric-path/switches", sn=self.sn)

    @property
    def cisco_fabric_path_routes(self):
        return models.Table(client=self.client, endpoint="tables/platforms/fabric-path/routes", sn=self.sn)

    @property
    def juniper_cluster(self):
        return models.Table(client=self.client, endpoint="tables/platforms/cluster/srx", sn=self.sn)

    @property
    def cisco_fex_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/platforms/fex/interfaces", sn=self.sn)

    @property
    def cisco_fex_modules(self):
        return models.Table(client=self.client, endpoint="tables/platforms/fex/modules", sn=self.sn)

    @property
    def cisco_vdc_devices(self):
        return models.Table(client=self.client, endpoint="tables/platforms/vdc/devices", sn=self.sn)

    @property
    def platform_cisco_vss(self):
        return models.Table(client=self.client, endpoint="tables/platforms/vss/overview", sn=self.sn)

    @property
    def cisco_vss_chassis(self):
        return models.Table(client=self.client, endpoint="tables/platforms/vss/chassis", sn=self.sn)

    @property
    def cisco_vss_vsl(self):
        return models.Table(client=self.client, endpoint="tables/platforms/vss/vsl", sn=self.sn)

    @property
    def poe_devices(self):
        return models.Table(client=self.client, endpoint="tables/platforms/poe/devices", sn=self.sn)

    @property
    def poe_interfaces(self):
        return models.Table(client=self.client, endpoint="tables/platforms/poe/interfaces", sn=self.sn)

    @property
    def poe_modules(self):
        return models.Table(client=self.client, endpoint="tables/platforms/poe/modules", sn=self.sn)

    @property
    def stacks(self):
        return models.Table(client=self.client, endpoint="tables/platforms/stack", sn=self.sn)

    @property
    def stacks_members(self):
        return models.Table(client=self.client, endpoint="tables/platforms/stack/members", sn=self.sn)

    @property
    def stacks_stack_ports(self):
        return models.Table(client=self.client, endpoint="tables/platforms/stack/connections", sn=self.sn)

    @property
    def logical_devices(self):
        return models.Table(client=self.client, endpoint="tables/platforms/devices", sn=self.sn)
